﻿Imports System.IO
Imports ICSharpCode.SharpZipLib.Core
Imports ICSharpCode.SharpZipLib.Zip
Imports System.Collections.Specialized

Public Class Backup

    Private pRemotePC As String             ' the NetBIOS name for remote operations (set to an empty string for local machine)
    Private pStartingDir As String          ' the starting location for backup source
    Private pArchivePath As String          ' the full path to the resulting archive
    Private pContinueOnErrors As Boolean    ' continue on errors (typically set to true)
    Private pSkipList As StringCollection   ' list of paths to skip (supports wild cards, ie: C:\Users\*\AppData\Local\Temp)
    Private pLogPath As String              ' the full path to an (optional) logfile
    Private pLogLevel As LoggingLevel       ' the level of detail in the logfile
    Private pFileCount As Integer           ' number of files copied
    Private pDirCount As Integer            ' number of directories copied

    Private log As StreamWriter
    Private zip As ZipOutputStream
    Private buffer As Byte() = New Byte(40959) {}

    Public Enum LoggingLevel As Integer
        None = 0
        Summary
        Errors
        Full
    End Enum

    ' the default constructor
    Public Sub New()
        pRemotePC = ""
        pStartingDir = ""
        pArchivePath = ""
        pContinueOnErrors = True
        pSkipList = New StringCollection
        pLogPath = ""
        pLogLevel = LoggingLevel.None
        pFileCount = 0
        pDirCount = 0
        log = Nothing
    End Sub

    ' the full constructor
    Public Sub New(ByVal Starting_Dir As String, ByVal Archive_Path As String, ByVal Continue_OnErrors As Boolean, ByVal Skip_list As StringCollection, ByVal Log_Path As String, ByVal Log_Level As LoggingLevel)
        pRemotePC = ""
        pStartingDir = Starting_Dir
        pArchivePath = Archive_Path
        pContinueOnErrors = Continue_OnErrors
        SkipList = Skip_list
        pLogPath = Log_Path
        pLogLevel = Log_Level
        pFileCount = 0
        pDirCount = 0
    End Sub

    ' the NetBIOS name for remote operations (set to an empty string for local machine)
    Public Property RemotePC() As String
        Get
            Return pRemotePC
        End Get
        Set(ByVal value As String)
            pRemotePC = value
        End Set
    End Property

    ' the starting location for backup source
    Public Property StartingDir() As String
        Get
            Return pStartingDir
        End Get
        Set(ByVal value As String)
            pStartingDir = value
        End Set
    End Property

    ' the full path to the resulting archive
    Public Property ArchivePath() As String
        Get
            Return pArchivePath
        End Get
        Set(ByVal value As String)
            pArchivePath = value
        End Set
    End Property

    ' continue on errors (typically set to true)
    Public Property ContinueOnErrors() As Boolean
        Get
            Return pContinueOnErrors
        End Get
        Set(ByVal value As Boolean)
            pContinueOnErrors = value
        End Set
    End Property

    ' list of paths to skip (supports wild cards, ie: C:\Users\*\AppData\Local\Temp)
    Public Property SkipList() As StringCollection
        Get
            Return pSkipList
        End Get
        Set(ByVal value As StringCollection)
            pSkipList = New StringCollection
            ' Format the "skip list"... 
            For Each skip As String In value
                If Not skip.EndsWith("\") Then
                    pSkipList.Add(skip.ToUpper & "\")
                Else
                    pSkipList.Add(skip.ToUpper)
                End If
            Next
        End Set
    End Property

    ' the full path to an (optional) logfile
    Public Property LogFile() As String
        Get
            Return pLogPath
        End Get
        Set(ByVal value As String)
            pLogPath = value
        End Set
    End Property

    ' the level of detail in the logfile
    Public Property LogLevel() As LoggingLevel
        Get
            Return pLogLevel
        End Get
        Set(ByVal value As LoggingLevel)
            pLogLevel = value
        End Set
    End Property

    ' number of files copied
    Public ReadOnly Property FileCount() As Integer
        Get
            Return pFileCount
        End Get
    End Property

    ' number of directories copied
    Public ReadOnly Property DirectoryCount() As Integer
        Get
            Return pDirCount
        End Get
    End Property

    ' perform the backup!
    Public Function PerformBackup() As Boolean
        Dim sp As New SetPrivileges
        Dim SkipListFormatted As New StringCollection
        pFileCount = 0
        pDirCount = 0

        ' set the backup privilege on both the local and remote PC
        ' TODO: Test need for both local and remote
        sp.SetPrivileges("")
        If pRemotePC <> "" Then
            sp.SetPrivileges(pRemotePC)
        End If

        ' setup logging
        ' TODO: Logging levels
        If pLogLevel <> LoggingLevel.None Then
            Try
                log = File.AppendText(pLogPath)
                log.WriteLine("Starting: " & Now)
                log.WriteLine("PC: " & IIf(pRemotePC = "", Environment.MachineName, pRemotePC).ToString)
            Catch ex As Exception
                Throw New ApplicationException("Can't open log file", ex)
            End Try
        End If

        ' open the zip archive file
        Try
            ' creates or truncates the zip file!
            zip = New ZipOutputStream(File.Create(pArchivePath))
            ' maximum compression
            zip.SetLevel(9)
            ' use "Zip 64" when appropriate
            zip.UseZip64 = UseZip64.Dynamic
        Catch ex As Exception
            Throw New ApplicationException("Can't open zip file", ex)
        End Try

        ' do it!
        Doit(ConvUNC(pStartingDir))

        ' clean up and go home
        If Not IsNothing(log) Then
            log.WriteLine("Files: " & pFileCount)
            log.WriteLine("Directories: " & pDirCount)
            log.WriteLine("Ending: " & Now)
            log.Close()
        End If
        zip.Close()

        Return True
    End Function

    ' traverse the directory structure
    Private Sub Doit(ByVal StartingDir As String)
        Dim ze As ZipEntry
        Dim pdi, di As DirectoryInfo
        Dim fi As FileInfo
        Dim temp As String

        ' dress it up a wee bit...
        If Not StartingDir.EndsWith("\") Then
            StartingDir = StartingDir & "\"
        End If

        ' skip directories that we don't care about...
        temp = UnConvUNC(StartingDir).ToUpper
        For Each skip As String In pSkipList
            If temp Like skip Or temp = skip Then
                'Debug.WriteLine("skipping: " & StartingDir)
                Exit Sub
            End If
        Next

        ' create an entry for this starting directory (otherwise you wouldn't get emptry directories)
        Try
            pdi = New DirectoryInfo(StartingDir)
            ze = New ZipEntry(UnConvUNC(StartingDir))
            ze.DateTime = pdi.LastWriteTime
            zip.PutNextEntry(ze)
        Catch ex As Exception
            If Not IsNothing(log) Then
                log.WriteLine(ex.Message)
            End If
            ' fatal (er, well... at least for this cycle)
            Exit Sub
        End Try

        ' do the files first
        For Each fi In pdi.GetFiles
            Try
                ' open the file for read access
                Using fs As FileStream = File.OpenRead(fi.FullName)
                    ze = New ZipEntry(UnConvUNC(fi.FullName))
                    ze.DateTime = fi.LastWriteTime
                    ze.ExternalFileAttributes = CInt(fi.Attributes)
                    zip.PutNextEntry(ze)
                    StreamUtils.Copy(fs, zip, buffer)
                End Using

                'Debug.WriteLine("fname=" & fname)
                pFileCount += 1
            Catch ex As Exception
                If Not IsNothing(log) Then
                    log.WriteLine(ex.Message)
                End If
                If Not ContinueOnErrors Then
                    Throw New ApplicationException("Could not continue", ex)
                End If
            End Try
        Next

        ' next, descend into the directory structure and start over
        Try
            For Each di In pdi.GetDirectories
                ' don't follow ReParse points
                If (di.Attributes And FileAttributes.ReparsePoint) = FileAttributes.ReparsePoint Then
                    Continue For
                End If

                ' recursive routine
                Doit(di.FullName)
                'Debug.WriteLine("dname=" & dname)
                pDirCount += 1
            Next
        Catch ex As Exception
            If Not IsNothing(log) Then
                log.WriteLine(StartingDir & ":" & ex.Message)
            End If
            If Not ContinueOnErrors Then
                Throw New ApplicationException("Could not continue", ex)
            End If
        End Try

    End Sub

    ' Convert a Remote PC name, Local path name pair into a UNC style path
    Private Function ConvUNC(ByVal Path As String) As String
        Dim sb As New System.Text.StringBuilder

        If pRemotePC = "" Or String.Compare(pRemotePC, Environment.MachineName, True) = 0 Then
            Return Path
        End If

        If Path.Length > 2 Then
            sb.Append("\\")
            sb.Append(pRemotePC)
            sb.Append("\")
            sb.Append(Path.Substring(0, 1))
            sb.Append("$")
            sb.Append(Path.Substring(2))
        End If

        Return sb.ToString
    End Function

    ' Converts a UNC path back to it's drive letter equivalent.
    Private Function UnConvUNC(ByVal Path As String) As String
        Dim sb As New System.Text.StringBuilder
        Dim index As Integer

        index = Path.IndexOf("$")
        If index > 0 Then
            sb.Append(Path.Substring(index - 1, 1).ToUpper)
            sb.Append(":")
            sb.Append(Path.Substring(index + 1))
        Else
            Return Path
        End If

        Return sb.ToString
    End Function
End Class
